# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

from __future__ import absolute_import
from os import path
import numpy as np
import itertools
try:
    from html.parser import HTMLParser
except:
    from HTMLParser import HTMLParser  # noqa

import pytest

import bioread
from bioread import reader
from bioread.reader import Reader
import logging
logger = reader.logger
logger.setLevel(logging.DEBUG)

DATA_PATH = path.join(path.dirname(path.abspath(__file__)), "data")

DATASETS = ['physio', 'nojournal']
# 3.8.1 is special, the compressed / uncompressed data don't agree in the last
# chunk. I think it's a bug in acqknowledge.
BADEND_VERSIONS = ['3.8.1']

# Since our data were backconverted to 3.8.1 and then up from there, they're
# all subject to the same bug that bit that version, and they're all different
# than 4.4.0
ORIG_VERSION = ['4.4.0']

# If you have more versions, send me updates!
NORMAL_VERSIONS = [
    '3.9.1',
    '4.1.0',
    '4.2.0',
    '4.3.0',
]

ALL_VERSIONS = BADEND_VERSIONS + NORMAL_VERSIONS + ORIG_VERSION

TEXT_JOURNAL_VERSIONS = [
    '3.8.1',
    '3.9.1',
    '4.1.0'
]

HTML_JOURNAL_VERSIONS = [
    '4.2.0',
    '4.3.0',
    '4.4.0',
    '5.0.1'
]

# Yes this is ridiculous but it helps
COMPRESSIONS = ['', '-c']


def data_file_names(datasets, versions, compressions):
    for dset, ver, comp in itertools.product(
            datasets, versions, compressions):
        yield data_file_name(dset, ver, comp)


def data_file_name(dataset, version, compression):
    return path.join(DATA_PATH, dataset, '{0}-{1}{2}.acq'.format(
        dataset, version, compression))


# Test to make sure we can read everything. This should throw an exception
# if things fail.
@pytest.mark.parametrize(
    'pathname',
    data_file_names(
        DATASETS,
        BADEND_VERSIONS + NORMAL_VERSIONS,
        COMPRESSIONS))
def test_reading(pathname):
    assert Reader.read(pathname), 'Error reading {0}'.format(pathname)


#@pytest.fixture(scope='module')
def canonical_files():
    versions = ['3.8.1', '4.1.0', '4.4.0']
    return dict(
        (
            (ver, dataset), Reader.read(
                data_file_name(dataset, ver, '-c')).datafile
        )
        for ver, dataset in itertools.product(versions, DATASETS)
    )


@pytest.mark.parametrize(
    'test_file,canon_data',
    itertools.chain(
        zip(
            data_file_names(['physio'], ALL_VERSIONS, COMPRESSIONS),
            itertools.repeat(canonical_files()['4.4.0', 'physio'])
        ),
        zip(
            data_file_names(['nojournal'], ALL_VERSIONS, COMPRESSIONS),
            itertools.repeat(canonical_files()['4.4.0', 'nojournal'])

        )
    )
)
def test_full_pattern_channels_match(test_file, canon_data):
    test_data = bioread.read(test_file)
    slices = full_pattern_slices(canon_data.channels)

    for cch, dch, s, i in zip(
            canon_data.channels,
            test_data.channels,
            slices,
            range(len(canon_data.channels))):
        result = np.array_equal(cch.raw_data[s], dch.raw_data[s])
        assert result, '{0} channel {1} does not match'.format(test_file, i)


@pytest.mark.parametrize(
    'dataset,version',
    itertools.product(DATASETS, NORMAL_VERSIONS + ORIG_VERSION)
)
def test_compresed_uncompressed_channels_match(dataset, version):
    uchan = bioread.read(data_file_name(dataset, version, '')).channels
    cchan = bioread.read(data_file_name(dataset, version, '-c')).channels
    for cch, uch, i in zip(uchan, cchan, range(len(uchan))):
        result = np.array_equal(cch.raw_data, uch.raw_data),
        assert result, 'Mismatch for {0}, {1} channel {2}'.format(
            dataset, version, i)


@pytest.mark.parametrize(
    'dataset,version',
    itertools.product(DATASETS, ALL_VERSIONS)
)
def test_compressed_uncompressed_markers_match(dataset, version):
    um = bioread.read_headers(
        data_file_name(dataset, version, '')).event_markers
    cm = bioread.read_headers(
        data_file_name(dataset, version, '-c')).event_markers
    assert um == cm


@pytest.mark.parametrize(
    'test_file,canon_data',
    zip(
        data_file_names(['physio'], HTML_JOURNAL_VERSIONS, COMPRESSIONS),
        itertools.repeat(canonical_files()['4.4.0', 'physio'])
    )
)
def test_html_journals_match(test_file, canon_data):
    test_data = bioread.read(test_file)
    canon_parser = DataExtractor()
    test_parser = DataExtractor()

    canon_parser.feed(canon_data.journal)
    test_parser.feed(test_data.journal)
    assert canon_parser.content == test_parser.content


@pytest.mark.parametrize(
    'test_file,canon_data',
    zip(
        data_file_names(['physio'], TEXT_JOURNAL_VERSIONS, COMPRESSIONS),
        itertools.repeat(canonical_files()['3.8.1', 'physio'])
    )
)
def test_text_journals_match(test_file, canon_data):
    test_data = bioread.read(test_file)

    test_journal = normalize_line_endings(test_data.journal)
    canon_journal = normalize_line_endings(canon_data.journal)

    assert test_journal == canon_journal

def test_reading_r35_file():
    filename = path.join(DATA_PATH, "misc", "r35_test.acq")
    test_data = bioread.read(filename)  # This will raise an exception on fail
    assert len(test_data.channels) == 2

def test_read_iso_8859_1():
    filename = path.join(DATA_PATH, "misc", "iso_8859_1.acq")
    test_data = bioread.read(filename)  # This will raise an exception on fail

    assert len(test_data.channels)==4



# This is kind of intense for something used by tests -- but the deal is:
# different versions of acqknowledge are treating the versions upconverted
# from 3.8.1 differently in the last, partially-filled pattern.
# The data should be identical between all files in a dataset for the
# filled-pattern parts. So this function returns a slice that'll get you that
# part of the raw data.
def full_pattern_slices(channels):
    pattern = reader.sample_pattern([c.frequency_divider for c in channels])
    point_counts = np.array([c.point_count for c in channels])
    pattern_uses = np.bincount(pattern)
    full_pattern_counts = point_counts - (point_counts % pattern_uses)
    return [slice(count) for count in full_pattern_counts]


def normalize_line_endings(s):
    return s.replace('\r\n', '\n').replace('\r', '\n')


# A little thing that'll let us strip text (in a horrible manner) from html
class DataExtractor(HTMLParser):
    def handle_data(self, data):
        if not hasattr(self, 'content'):
            self.content = ''
        self.content += data


# Lower-level function tests.
def test_greatest_common_denominator():
    assert reader.greatest_common_denominator(8, 12) == 4
    assert reader.greatest_common_denominator(0, 8) == 8


def test_least_common_multiple():
    assert reader.least_common_multiple(4) == 4
    assert reader.least_common_multiple(2, 8) == 8
    assert reader.least_common_multiple(8, 2) == 8
    assert reader.least_common_multiple(2, 7) == 14
    assert reader.least_common_multiple(2, 3, 8) == 24


def assert_pattern(dividers, pattern):
    assert np.array_equal(reader.sample_pattern(dividers), pattern)


def test_sample_pattern():
    assert_pattern([1], [0])
    assert_pattern([1, 2], [0, 1, 0])
    assert_pattern([2, 2], [0, 1])
    assert_pattern([1, 4, 2], [0, 1, 2, 0, 0, 2, 0])



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread
#
# This is a very simple thing to write an AcqKnowledge file as tab-delimited
# text. It'll produce very large output, so use it with care.

from __future__ import unicode_literals
import csv


def write_text(datafile, out_stream, channel_indexes, missing_val):
    writer = csv.writer(out_stream, delimiter=str("\t"))
    if not channel_indexes:
        channel_indexes = range(len(datafile.channels))
    chans = [datafile.channels[i] for i in channel_indexes]
    headers = ["time (s)"] + [
        "{0} ({1})".format(c.name, c.units) for c in chans]
    writer.writerow(headers)
    for i, t in enumerate(datafile.time_index):
        rd = [t] + [data_or_blank(c, i, missing_val) for c in chans]
        writer.writerow(rd)


def data_or_blank(channel, index, missing_val):
    ci = index // channel.frequency_divider
    if index % channel.frequency_divider == 0 and ci < channel.point_count:
        return channel.data[ci]
    return missing_val



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# NOTE: This file must not import anything, or it will break installation.

version_tuple = (3, 0, 2)
version = ".".join([str(p) for p in version_tuple])
version_description = "bioread {0}".format(version)

author = "Original author: Nate Vack, current maintainer: Meng"
author_email = "Original author: njvack@wisc.edu, current maintainer: Meng"
license = "MIT"
copyright = (
    "Copyright (c) 2023 Boards of Regent of the University of Wisconsin System"  # noqa
)
url = "https://github.com/Schweik7/bioread"



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

from __future__ import division
import numpy as np
from datetime import datetime, timezone, timedelta

REF_DATE = datetime(1970, 1, 1, tzinfo=timezone.utc)


class Datafile(object):
    """
    A data file for the AcqKnowledge system. Generally, gonna be created
    from a file by reader.Reader.
    """

    def __init__(
            self,
            graph_header=None, channel_headers=None, foreign_header=None,
            channel_dtype_headers=None, samples_per_second=None, name=None,
            marker_header=None, marker_item_headers=None,):
        self.graph_header = graph_header
        self.channel_headers = channel_headers
        self.foreign_header = foreign_header
        self.channel_dtype_headers = channel_dtype_headers
        self.samples_per_second = samples_per_second
        self.name = name
        self.marker_header = marker_header
        self.marker_item_headers = marker_item_headers
        self.event_markers = None
        self.journal_header = None
        self.journal = None
        self.__named_channels = None
        self.__time_index = None

        self.channels = self.__build_channels()
        self.channel_order_map = dict(
            [[c.order_num, c] for c in self.channels]
        )

    @property
    def named_channels(self):
        if self.__named_channels is None and self.channels is not None:
            self.__named_channels = {}
            for c in self.channels:
                self.__named_channels[c.name] = c

        return self.__named_channels

    @property
    def is_compressed(self):
        return self.graph_header.compressed

    @property
    def data_length(self):
        if self.is_compressed:
            return 0
        return sum([c.data_length for c in self.channels])

    def __str__(self):
        return("AcqKnowledge file (rev %s): %s channels, %s samples/sec" % (
            self.graph_header.file_revision, self.graph_header.channel_count,
            self.samples_per_second))

    def __repr__(self):
        return str(self)

    @property
    def time_index(self):
        if self.__time_index is not None:
            return self.__time_index

        total_samples = max(
            [ch.frequency_divider * ch.point_count
                for ch in self.channel_headers])
        total_seconds = total_samples / self.samples_per_second
        self.__time_index = np.linspace(0, total_seconds, total_samples)
        return self.__time_index

    @property
    def earliest_marker_created_at(self):
        """
        Guess date at which the file was created by finding the
        minimum date_created_utc from the event markers, or return
        None if there are no markers our markers don't
        have date_created_utc
        """
        if self.event_markers is None:
            return None
        created_times = [
            m.date_created_utc for m in self.event_markers
            if m.date_created_utc]
        try:
            return min(created_times)
        except (ValueError, TypeError):
            return None

    def __build_channels(self):
        return [
            Channel.from_headers(
                chan_hdr=ch,
                dtype_hdr=cdh,
                samples_per_second=self.samples_per_second,
                datafile=self)
            for ch, cdh in zip(
                self.channel_headers, self.channel_dtype_headers)
        ]

    def __set_channel_time_indexes(self):
        for c in self.channels:
            c.time_index = self.time_index[::c.frequency_divider]


class Channel(object):
    """
    An individual channel of Biopac data. Has methods to access raw data from
    the file, as well as a scaled copy if the raw data is in integer format.
    Also generally created by reader.Reader.
    """

    def __init__(
            self,
            frequency_divider=None, raw_scale_factor=None, raw_offset=None,
            name=None, units=None, fmt_str=None, samples_per_second=None,
            point_count=None, order_num=None, datafile=None):

        self.frequency_divider = frequency_divider
        self.raw_scale_factor = raw_scale_factor
        self.raw_offset = raw_offset
        self.name = name
        self.units = units
        self.fmt_str = fmt_str
        self.samples_per_second = samples_per_second
        self.point_count = point_count
        self.dtype = np.dtype(fmt_str)
        # For some reason, scale and offset lie for float files
        if self.dtype.kind == 'f':
            self.raw_scale_factor = 1
            self.raw_offset = 0
        self.order_num = order_num
        self.datafile = datafile

        # Don't allocate storage automatically -- this means we can read
        # only some channels or stream the data without putting all the data
        # in memory.
        self.raw_data = None
        self.__data = None
        self.__upsampled_data = None

    @classmethod
    def from_headers(
            cls,
            chan_hdr,
            dtype_hdr,
            samples_per_second,
            datafile=None):
        divider = chan_hdr.frequency_divider
        chan_samp_per_sec = samples_per_second / divider

        return cls(
            frequency_divider=divider,
            raw_scale_factor=chan_hdr.raw_scale,
            raw_offset=chan_hdr.raw_offset,
            name=chan_hdr.name,
            units=chan_hdr.units,
            fmt_str=dtype_hdr.numpy_dtype,
            samples_per_second=chan_samp_per_sec,
            point_count=chan_hdr.point_count,
            order_num=chan_hdr.order_num,
            datafile=datafile
        )

    def _allocate_raw_data(self):
        self.raw_data = np.zeros(self.point_count, dtype=self.dtype)

    @property
    def sample_size(self):
        """
        The size, in bytes, of one sample's worth of data.
        """
        return self.dtype.itemsize

    @property
    def data_length(self):
        """
        The size, in bytes, of the entire channel's raw data stream.
        """
        return self.sample_size * self.point_count

    @property
    def loaded(self):
        return self.raw_data is not None

    @property
    def time_index(self):
        if self.datafile is None:
            return None
        div = self.frequency_divider
        return self.datafile.time_index[::div][0:self.point_count]

    @property
    def data(self):
        """
        The channel's data, scaled by the raw_scale_factor and offset. These
        will be the values reported by AcqKnowledge. Note: only integer data
        types are scaled and offset.
        """
        if not self.loaded:
            return None
        if self.__data is not None:
            return self.__data
        if self.dtype.kind == 'f':
            self.__data = self.raw_data
        else:
            self.__data = (
                (self.raw_data * self.raw_scale_factor) + self.raw_offset)
        return self.__data

    @property
    def upsampled_data(self):
        """
        The channel's data, sampled at the native frequency of the file.
        All channels will have the same number of points using this method,
        unless recording stopped in the middle of a block.
        Nearest-neighbor sampling is used.
        """
        if self.__upsampled_data is None:
            total_samples = self.data.shape[0]*self.frequency_divider
            self.__upsampled_data = self.data[
                np.arange(total_samples)//self.frequency_divider]
        return self.__upsampled_data

    def free_data(self):
        self.raw_data = None
        self.__data = None
        self.__upsampled_data = None

    def __str__(self):
        return("Channel %s: %s samples, %s samples/sec, loaded: %s" % (
            self.name,
            self.point_count,
            self.samples_per_second,
            self.loaded))

    def __repr__(self):
        return str(self)


class EventMarker(object):
    """
    A marker -- some kind of annotation for an AcqKnowledge file. They all
    have a sample index and some text, and more modern ones can be one of
    several styles (say, a flag or a star or a waveform start) and can be
    attached to a particular channel.
    """

    def __init__(
            self,
            time_index,
            sample_index,
            text,
            channel_number,
            channel=None,
            date_created_ms=None,
            type_code=None):

        self.time_index = time_index
        self.sample_index = sample_index
        self.text = text
        self.channel_number = channel_number
        self.channel = channel
        self.date_created_utc = None
        try:
            if date_created_ms is not None:
                self.date_created_utc = (
                    REF_DATE + timedelta(milliseconds=date_created_ms)
                )
        except OverflowError:
            # Sometimes the dates look to be insanely huge, I don't know
            # what's up with this, honestly.
            pass  # for now we'll just leave date_created_utc as None

        self.type_code = type_code
        super(EventMarker, self).__init__()

    def __eq__(self, other):
        return all([
            self.sample_index == other.sample_index,
            self.text == other.text,
            self.channel_number == other.channel_number,
            self.date_created_utc == other.date_created_utc,
            self.type_code == other.type_code
        ])

    def __str__(self):
        return("EventMarker {}: idx: {} channel: {} date_created_utc: {} type_code: {}".format(  # noqa: E501
            self.text,
            self.sample_index,
            self.channel_number,
            self.date_created_utc.isoformat(),
            self.type_code))

    def __repr__(self):
        return str(self)

    @property
    def type(self):
        if self.type_code is None:
            return "None"
        return MARKER_TYPE_MAP.get(self.type_code, "Unknown")

    @property
    def channel_sample_index(self):
        if self.channel is None:
            return None
        return self.sample_index // self.channel.frequency_divider

    @property
    def channel_name(self):
        if self.channel is None:
            return None
        return self.channel.name

    @property
    def date_created_str(self):
        if hasattr(self.date_created_utc, 'isoformat'):
            return self.date_created_utc.isoformat()
        return 'Unknown'


MARKER_TYPE_MAP = {
    'apnd': 'Append',
    'defl': 'Default',
    'wfon': 'Waveform Onset',
    'wfof': 'Waveform End',
    'nois': 'Change in Signal Quality',
    'rhyt': 'Change in Rhythm',
    'recv': 'Recovery',
    'max ': 'Maximum',
    'min ': 'Minimum',
    'rset': 'Reset',
    'cmlb': 'Communication Lost Begin',
    'cmle': 'Communication Lost End',
    'ansh': 'Short Arrow',
    'anmd': 'Medium Arrow',
    'anlg': 'Long Arrow',
    'flag': 'Flag',
    'star': 'Star',
    'usr1': 'User Type 1',
    'usr2': 'User Type 2',
    'usr3': 'User Type 3',
    'usr4': 'User Type 4',
    'usr5': 'User Type 5',
    'usr6': 'User Type 6',
    'usr7': 'User Type 7',
    'usr8': 'User Type 8',
    'usr9': 'User Type 9',
    'qrsb': 'QRS Onset',
    'qrs ': 'QRS Peak',
    'qrse': 'QRS End',
    'tbeg': 'T-wave Onset',
    't   ': 'T-wave Peak',
    'tend': 'T-wave End',
    'pbeg': 'P-wave Onset',
    'p   ': 'P-wave Peak',
    'pend': 'P-wave End',
    'q   ': 'Q-wave Peak',
    's   ': 'S-wave Peak',
    'u   ': 'U-wave Peak',
    'pq  ': 'PQ Junction',
    'jpt ': 'J-point',
    'stch': 'ST Segment Change',
    'tch ': 'T-wave Change',
    'nrml': 'Normal Beat',
    'pace': 'Paced Beat',
    'pfus': 'Fusion of Paced and Normal Beat',
    'lbbb': 'Left Bundle Branch Block Beat',
    'rbbb': 'Right Bundle Branch Block Beat',
    'bbb ': 'Bundle Branch Block Beat',
    'apc ': 'Atrial Premature Beat',
    'aber': 'Aberrated Atrial Prematuire Beat',
    'npc ': 'Nodal Premature Beat',
    'svpb': 'Supraventricular Premature Beat',
    'pvc ': 'Premature Ventricular Contraction',
    'ront': 'R-on-T Premature Ventricular Contraction',
    'fusi': 'Fusion of Ventricular and Normal Beat',
    'aesc': 'Atrial Escape Beat',
    'nesc': 'Nodal Escape Beat',
    'sves': 'Supraventricular Escape Beat',
    'vesc': 'Ventricular Escape Beat',
    'syst': 'Systole',
    'dias': 'Diastole',
    'edp ': 'End Diastolic Pressure',
    'aptz': 'A-point',
    'bptz': 'B-point',
    'cptz': 'C-point',
    'xptz': 'X-point',
    'yptz': 'Y-point',
    'optz': 'O-point',
    'plat': 'Plateau',
    'upst': 'Upstroke',
    'vfon': 'Start of Ventricular Flutter',
    'flwa': 'Ventricular Flutter Wave',
    'vfof': 'End of Ventricular Flutter',
    'pesp': 'Pacemaker Artifact',
    'arfc': 'Isolated QRS-like Artifact',
    'napc': 'Non-conducted P-wave',
    'base': 'Baseline',
    'dose': 'Dose',
    'wash': 'Wash',
    'apon': 'Spike Episode Begin',
    'apof': 'Spike Episode End',
    'rein': 'Inspire Start',
    'reot': 'Expire Start',
    'reap': 'Apnea Start',
    'stim': 'Stimulus Delivery',
    'resp': 'Response',
    'scr ': 'Skin Conductance Response',
    'sscr': 'Specific SCR',
    'ctr1': 'Cluster 1',
    'ctr2': 'Cluster 2',
    'ctr3': 'Cluster 3',
    'ctr4': 'Cluster 4',
    'ctr5': 'Cluster 5',
    'ctr6': 'Cluster 6',
    'ctr7': 'Cluster 7',
    'ctr8': 'Cluster 8',
    'ctr9': 'Cluster 9',
    'ctrn': 'Cluster n',
    'cend': 'End Cluster',
    'outl': 'Outlier',
    'tran': 'Training Set',
    'cut ': 'Cut',
    'vb  ': 'Paste Begin',
    've  ': 'Paste End',
    'selb': 'Selection Begin',
    'sele': 'Selection End',
    'steb': 'Start of Eye Blink Artifact',
    'eneb': 'End of Eye Blink Artifact',
    'sexc': 'Start of Excursion Artifact',
    'eexc': 'End of Excursion Artifact',
    'ssat': 'Start of Saturation Artifact',
    'esat': 'End of Saturation Artifact',
    'sspk': 'Start of Spike Artifact',
    'espk': 'End of Spike Artifact',
    'semg': 'Start of EMG Artifact',
    'eemg': 'End of EMG Artifact',
    'wles': 'Workload - EMG Start',
    'wlee': 'Workload - EMG End',
    'ipss': 'Workload - Invalid PSD Start',
    'ipse': 'Workload - Invalid PSD End',
    'ddst': 'Dummy Data Start',
    'dded': 'Dummy Data End',
    'idst': 'Misaligned Data',
    'bprs': 'Button Pressed',
    'leho': 'Left Eye Hit Object',
    'reho': 'Right Eye Hit Object',
    'smis': 'SMI Stimulus Image Has Been Presented to the Subject',
    'mors': 'Start Out of Range',
    'more': 'End Out of Range'
}



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# This is an enumeration of known file revisions, for use in header parsing.

V_ALL = 0
V_20a = 30
V_20b = 31
V_20r = 32
V_207 = 33
V_30r = 34
V_303 = 35
V_35x = 36
V_36x = 37
V_370 = 38
V_373 = 39
V_381 = 41
V_37P = 42
V_382 = 43
V_38P = 44
V_390 = 45
V_400B = 61  # Earliest 'late version' one I've seen so far...
V_400 = 68
V_401 = 76
V_402 = 78  # Unsure about exact versions... but this is Post-Some-Big-Change.
V_41a = 80  # But I'm honestly guessing about some of these.
V_410 = 83
V_411 = 84
V_420 = 108
V_42x = 121  # Not yet sure what version this is
V_430 = 124
V_440 = 128
V_501 = 132



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# flake8: noqa: E741

# Note that the concept of this is fundamentally silly, I'd use ctypes if I were
# going ot build something like this again. But I didn't know about ctypes when
# I started.

from __future__ import absolute_import

import struct


class StructDict(object):
    """
    TODO: This class is basically a shittier version of ctypes. Remove this
          and replace it with ctypes.

    This class allows you to declare a binary file's header structure with
    name and type information, and then will unpack the header into a
    dictionary.
    For example:
    >>> header_structure = [
        ('version', 'h'), ('xy_dim', '2b'), ('name', '5s')
    ]
    >>> sd = StructDict('>', header_structure)
    >>> sd.format_string
    'h2b5s'
    >>> sd.len_bytes
    9
    >>> header_data = '\x00\x01\x05\x10foo\x00\x00'
    >>> sd.unpack(header_data)
    {
        'version' : 1,     # Single-length elements are de-tupelized
        'xy_dim' : (5, 16),
        'name' : 'foo'     # Strings get trailing nulls trimmed
    }
    """

    def __init__(self, byte_order_char, struct_info=None):
        self.byte_order_char = byte_order_char
        self.struct_info = struct_info
        self.full_struct_info = None

    def unpack(self, data):
        """
        Return a dict with the unpacked data.
        """
        self.__setup()
        unpacked = struct.unpack(self.format_string, data)
        output = {}
        for name, fs, start_index, end_index in self.full_struct_info:
            l = end_index-start_index
            if l == 1:
                val = unpacked[start_index]
            else:
                val = unpacked[start_index:end_index]
            if type(val) == str:
                null_idx = val.find("\x00")
                if null_idx == -1:
                    null_idx = len(val)
                val = val[0:null_idx]
            output[name] = val
        return output

    def labeled_offsets_lengths(self):
        """
        Primarily for debugging purposes: generate a list of byte offsets
        and struct lengths, so you can see what fields are where. Then you
        can explore with your hex editor, or compare against spec, or
        whatever.

        Example:

        >>> sd = StructDict('>', [('version', 'h'), ('header_len', 'l')])
        >>> sd.labeled_offsete_lengths()
        [
            ('version', 'h', 0, 2),
            ('header_len', 'l', 2, 4)
        ]

        """
        table = []
        build_fs = self.byte_order_char
        for si in self.struct_info:
            name, fs = si[0:2]
            f_offset = struct.calcsize(build_fs)
            f_len = struct.calcsize(self.byte_order_char+fs)
            build_fs += fs
            table.append((name, fs, f_offset, f_len))

        return table

    @property
    def len_bytes(self):
        return struct.calcsize(self.format_string)

    @property
    def len_elements(self):
        return len(self.struct_info)

    @property
    def format_string(self):
        s = ''.join([si[1] for si in self.struct_info])
        return self.__bof_fs(s)

    def __setup(self):
        if self.full_struct_info is None:
            self.full_struct_info = self.__full_struct_info()

    def __bof_fs(self, format_str):
        return self.byte_order_char + format_str

    def __unpacked_element_count(self, format_str):
        # We need to figure this out by actually faking some data and
        # unpacking it. Crazy, huh?
        f_str = self.__bof_fs(format_str)
        f_len = struct.calcsize(f_str)
        dummy = struct.pack('%ss' % f_len, b'')
        unpacked = struct.unpack(f_str, dummy)
        return len(unpacked)  # The number of elements in the tuple

    def __full_struct_info(self):
        full_struct_info = []
        start_index = 0
        end_index = 0
        for si in self.struct_info:
            name, fs = si[0:2]
            tup_len = self.__unpacked_element_count(fs)
            end_index = start_index + tup_len
            full_struct_info.append((name, fs, start_index, end_index))
            start_index = end_index
        return full_struct_info



#!/usr/bin/env python
# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread
#
# This script writes an AcqKnowledge file as HDF5, using h5py:
# http://www.h5py.org/

"""Convert an AcqKnowledge file to an HDF5 file.

Usage:
  acq2hdf5 [options] <acq_file> <hdf5_file>
  acq2hdf5 -h | --help
  acq2hdf5 --version

Options:
  --values-as=<type>    Save raw measurement values, stored as integers in the
                        base file, as either 'raw' or 'scaled'. If stored as
                        raw, you can convert to scaled using the scale and
                        offset attributes on the channel. If storing scaled
                        values, scale and offset will be 1 and 0.
                        [default: scaled]
  --compress=<method>   How to compress data. Options are gzip, lzf, none.
                        [default: gzip]
  --data-only           Only save data and required headers -- do not save
                        journal or marker information.
  -v, --verbose         Print extra messages for debugging.
"""

import sys
import logging

from bioread import _metadata as meta
from bioread import reader as br
from docopt import docopt

# Re-adding the handler on reload causes duplicate log messages.
logger = logging.getLogger("bioread.runners.acq2hdf5")
logger.setLevel(logging.INFO)
log_handler = logging.StreamHandler()
log_handler.setLevel(logging.DEBUG)
log_handler.setFormatter(logging.Formatter("%(message)s"))

try:
    import h5py
except ImportError:
    logger.error("acq2hdf5 requires h5py")
    sys.exit(1)


COMPRESSION_OPTS = {
    'none': {},
    'gzip': {'compression': 'gzip'},
    'lzf': {'compression': 'lzf'}
}

SCALE_DATA = {
    'scaled': True,
    'raw': False
}


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    pargs = docopt(__doc__, argv, version=meta.version_description)
    if pargs['--verbose']:
        logger.setLevel(logging.DEBUG)
    br.logger.setLevel(logger.level)
    logger.debug(pargs)
    try:
        comp_opts = COMPRESSION_OPTS[pargs['--compress']]
    except KeyError:
        logger.error("Unknown compression: {0}".format(pargs['--compress']))
        sys.exit(1)
    scale = False
    try:
        scale = SCALE_DATA[pargs['--values-as']]
    except KeyError:
        logger.error("Unknown values-as option: {}".format(
            pargs['--values-as']))
    make_hdf5(
        pargs['<acq_file>'],
        pargs['<hdf5_file>'],
        comp_opts,
        scale,
        pargs['--data-only']
    )


def make_hdf5(
        acq_filename,
        hdf5_filename,
        compression_opts,
        scale,
        data_only):
    acq_file = open(acq_filename, 'rb')
    hdf5_file = h5py.File(hdf5_filename, 'w')
    r = br.Reader.read_headers(acq_file)
    df = r.datafile
    hdf5_file.attrs['acq_revision'] = r.file_revision
    hdf5_file.attrs['samples_per_second'] = r.samples_per_second
    hdf5_file.attrs['byte_order'] = r.byte_order_char
    if not data_only:
        hdf5_file.attrs['journal'] = df.journal or ''
    channel_datasets = None
    if r.is_compressed:
        channel_datasets = save_channels_compressed(
            acq_file, r, hdf5_file, compression_opts, scale)
    else:
        channel_datasets = save_channels_uncompressed(
            acq_file, r, hdf5_file, compression_opts, scale)
    channel_map = dict(
        [[ch.order_num, channel_datasets[i]]
            for i, ch in enumerate(df.channels)])
    if not data_only:
        save_markers(hdf5_file, df, channel_map)


def cnum_formatter(channel_count):
    return "channel_{{:0{0}d}}".format(len(str(channel_count)))


def create_channel_datasets(grp, channels, compression_opts, scale):
    cfstr = cnum_formatter(len(channels))

    channel_dsets = []
    for i, c in enumerate(channels):
        if scale:
            dset_kwargs = {'dtype': 'f8'}
        else:
            dset_kwargs = {'dtype': c.dtype}

        dset_kwargs.update(compression_opts)
        cnum_name = cfstr.format(i)
        logger.debug("Creating dataset {0}".format(cnum_name))
        dset = grp.create_dataset(
            cnum_name,
            (c.point_count,),
            **dset_kwargs)
        if c.dtype.kind == 'i' and not scale:
            dset.attrs['scale'] = c.raw_scale_factor
            dset.attrs['offset'] = c.raw_offset
        else:
            dset.attrs['scale'] = 1.0
            dset.attrs['offset'] = 0.0
        dset.attrs['name'] = c.name
        dset.attrs['frequency_divider'] = c.frequency_divider
        dset.attrs['units'] = c.units
        dset.attrs['samples_per_second'] = c.samples_per_second
        dset.attrs['channel_number'] = c.order_num
        channel_dsets.append(dset)
    return channel_dsets


def save_channels_uncompressed(
        acq_file,
        reader,
        hdf5_file,
        compression_opts,
        scale):
    logger.debug("Saving uncompressed data to hdf5")
    cg = hdf5_file.create_group('/channels')
    df = reader.datafile
    chunker = reader.stream()
    channel_dsets = create_channel_datasets(
        cg, df.channels, compression_opts, scale)

    for chunk_num, chunk_buffers in enumerate(chunker):
        logger.debug("Got chunk {0}".format(chunk_num))
        for buf, dset in zip(chunk_buffers, channel_dsets):
            if scale:
                chan = buf.channel
                dset[buf.channel_slice] = (
                    (buf.buffer[:] * chan.raw_scale_factor) +
                    chan.raw_offset)
            else:
                dset[buf.channel_slice] = buf.buffer[:]
    return channel_dsets


def save_channels_compressed(
        acq_file,
        reader,
        hdf5_file,
        compression_opts,
        scale):
    logger.debug("Saving compressed data to hdf5")
    cg = hdf5_file.create_group('/channels')
    df = reader.datafile
    channel_dsets = create_channel_datasets(
        cg, df.channels, compression_opts, scale)
    for i, (c, dset) in enumerate(zip(df.channels, channel_dsets)):
        logger.debug("Saving channel {0}".format(i))
        # This magically populates c.raw_data; I know this is kind of bad
        reader._read_data(channel_indexes=[i])
        if scale:
            dset[:] = (c.raw_data[:] * c.raw_scale_factor) + c.raw_offset
        else:
            dset[:] = c.raw_data[:]
        # Release the channel's memory
        c.free_data()
    return channel_dsets


def save_markers(hdf5_file, datafile, dset_map):
    markers = datafile.event_markers
    mcount = len(markers)
    marker_name_formatter = "marker_{{:0{0}d}}".format(len(str(mcount)))
    for i, m in enumerate(markers):
        mname = marker_name_formatter.format(i)
        mg = hdf5_file.create_group("/event_markers/{0}".format(mname))
        mg.attrs['label'] = m.text
        mg.attrs['global_sample_index'] = m.sample_index
        if m.type_code:
            mg.attrs['type_code'] = m.type_code
            mg.attrs['type'] = m.type
        mg.attrs['date_created'] = m.date_created_str
        if m.channel:
            mg.attrs['channel_number'] = m.channel_number
            cdset = dset_map[m.channel_number]
            mg['channel'] = cdset
            mg.attrs['channel_sample_index'] = (
                m.sample_index // cdset.attrs['frequency_divider'])


if __name__ == '__main__':
    argv = sys.argv[1:]
    main(argv)



#!/usr/bin/env python
# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# This contains the entry point for an executable to convert BIOPAC
# AcqKnowledge files into Matlab files.

"""Convert an AcqKnowledge file to a MATLAB file.

Usage:
  acq2mat [options] <acq_file> <mat_file>
  acq2mat -h | --help
  acq2mat --version

Options:
  -c, --compress  Save compressed Matlab file
  --data-only     Only save data and required header information -- do not
                  save event markers.

Note: scipy is required for this program.
"""

from __future__ import absolute_import

import sys
from docopt import docopt

from bioread.reader import Reader
from bioread.writers.matlabwriter import MatlabWriter
from bioread import _metadata as meta


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    amr = AcqToMatRunner(argv)
    amr.run()


class AcqToMatRunner(object):
    """The little wrapper class that converts acq files to mat files"""

    def __init__(self, argv, err=None):
        self.argv = argv
        if err is None:
            err = sys.stderr
        self.err = err

    def run(self):
        old_err = sys.stderr
        sys.stderr = self.err
        pargs = docopt(
            __doc__,
            self.argv,
            version=meta.version_description)
        try:
            import scipy  # noqa -- catch this error before matlabwriter
        except Exception:
            sys.stderr.write("scipy is required for writing matlab files\n")
            sys.exit(1)
        infile = pargs['<acq_file>']
        matfile = pargs['<mat_file>']
        compress = pargs['--compress']

        data = Reader.read(infile).datafile
        MatlabWriter.write_file(
            data,
            matfile,
            compress=compress,
            data_only=pargs['--data-only'])

        sys.stderr = old_err


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# This contains the entry point for an executable to convert BIOPAC
# AcqKnowledge files into text files
"""Write the data from an AcqKnowledge file channel to a text file.

Usage:
  acq2txt [options] <acq_file>
  acq2txt -h | --help
  acq2txt --version

Options:
  --version                    Show program's version number and exit.
  -h, --help                   Show this help message and exit.
  --channel-indexes=<indexes>  The indexes of the channels to extract.
                               Separate numbers with commas. Default is to
                               extract all channels.
  -o, --outfile=<file>         Write to a file instead of standard out.
  --missing-as=<val>           What value to write where a channel is not
                               sampled. [default: ]

The first column will always be time in seconds. Channel raw values are
converted with scale and offset into native units.
"""

import sys

from docopt import docopt

import bioread
from bioread.writers import txtwriter
from bioread import _metadata as meta


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    amr = AcqToTxtRunner(argv)
    amr.run()


class AcqToTxtRunner(object):
    """The little wrapper class that converts acq files to text files"""

    def __init__(self, argv):
        self.argv = argv

    def run(self):
        pargs = docopt(
            __doc__,
            self.argv,
            version=meta.version_description)
        infile = pargs['<acq_file>']
        channel_indexes = None
        if pargs['--channel-indexes']:
            channel_indexes = [
                int(i) for i in pargs['--channel-indexes'].split(',')]
        data = bioread.read(infile, channel_indexes=channel_indexes)
        mval = pargs['--missing-as']
        if pargs['--outfile']:
            with open(pargs['--outfile'], 'w', newline='') as f:
                txtwriter.write_text(data, f, channel_indexes, mval)
        else:
            txtwriter.write_text(data, sys.stdout, channel_indexes, mval)


if __name__ == '__main__':
    main()



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

from __future__ import absolute_import

from bioread import reader

from ._metadata import version as __version__, author as __author__  # noqa


def read(filelike, channel_indexes=None):
    """
    Read a file (either an IO object or a filename) and return a Datafile.

    channel_indexes:    A list of integer channel numbers. Other channels will
                        have empty data.
    target_chunk_size:  A guide for the number of bytes to read at a time.
    """
    return reader.Reader.read(filelike, channel_indexes).datafile


# Deprecated; provided for compatibility with previous versions.
read_file = read


def read_headers(filelike):
    """
    Read only the headers of a file, returns a Datafile with empty channels.
    """
    return reader.Reader.read_headers(filelike).datafile


def reader_for_streaming(io):
    """
    Read the headers of a file, return a Reader object that will allow you to
    stream the data in chunks with stream().
    """
    if not hasattr(io, 'read'):
        raise TypeError('{0} must be an opened file.'.format(io))
    if hasattr(io, 'encoding'):
        raise TypeError('{0} must be opened in binary mode'.format(io))
    return reader.Reader.read_headers(io)



#!/usr/bin/env python
# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# This contains the entry point for an executable to print basic information
# about an AcqKnowledge file.

"""Print some information about an AcqKnowledge file.

Usage:
    acq_info [options] <acq_file>
    acq_info -h | --help
    acq_info --version

Options:
  -d, --debug  print lots of debugging data

Note: Using - for <acq_file> reads from stdin.

"""

import sys
import logging

from io import BytesIO
from docopt import docopt

from bioread import reader

from bioread.reader import Reader
from bioread import _metadata as meta


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    air = AcqInfoRunner(argv)
    air.run()


class AcqInfoRunner(object):

    def __init__(self, argv, out=None, err=None):
        self.argv = argv
        if out is None:
            out = sys.stdout
        self.out = out
        if err is None:
            err = sys.stderr
        self.err = err

    def run(self):
        old_out = sys.stdout
        old_err = sys.stderr
        sys.stdout = self.out
        sys.stderr = self.err

        pargs = docopt(
            __doc__,
            self.argv,
            version=meta.version_description)
        if pargs['--debug']:
          reader.logger.level = logging.DEBUG

        df = None
        infile = pargs['<acq_file>']
        try:
            if infile == '-':
                df = BytesIO(sys.stdin.read())
            else:
                df = open(infile, 'rb')
        except Exception:
            sys.stderr.write("Error reading {0}\n".format(infile))
            sys.exit(1)

        self.reader = Reader(df)
        self.reader._read_headers()
        try:
            pass
        except Exception as e:
            sys.stderr.write("Error reading headers!\n")
            sys.stderr.write("File position: %s\n" % self.reader.acq_file.tell())
            sys.stderr.write(str(e))
            # Don't exit here; it'll still print what it can.

        if pargs['--debug']:
            self.__print_debug()
        else:
            self.__print_simple()

        sys.stderr = old_err
        sys.stdout = old_out

    def __print_simple(self):
        gh = self.reader.graph_header
        chs = self.reader.channel_headers
        cdhs = self.reader.channel_dtype_headers
        print("File revision: %s" % gh.file_revision)
        if self.reader.datafile.earliest_marker_created_at is not None:
            print(
                "Earliest event marker created at: %s" %
                self.reader.datafile.earliest_marker_created_at.isoformat()
            )
        print("Sample time: %s" % gh.sample_time)
        print("Compressed: %s" % gh.compressed)
        print("Number of channels: %s" % gh.channel_count)
        for ch, cdh in zip(chs, cdhs):
            print("%s:" % ch.name)
            print("\tUnits: %s" % ch.units)
            print("\tNumber of samples: %s" % ch.point_count)
            print("\tFrequency divider: %s" % ch.frequency_divider)

    def __print_debug(self):
        gh = self.reader.graph_header
        fh = self.reader.foreign_header
        chs = self.reader.channel_headers
        cdhs = self.reader.channel_dtype_headers

        print("Graph header starts at offset %s" % gh.offset)
        print(gh.data)
        print("\n")
        for i, ch in enumerate(chs):
            print("Channel header %s starts at offset %s" % (i, ch.offset))
            print(ch.data)
            print("\n")
        print("Foreign header starts at offset %s" % fh.offset)
        print(fh.data)
        print("\n")
        for i, cdh in enumerate(cdhs):
            print(
                "Channel dtype header %s starts at offset %s" %
                (i, cdh.offset))
            print(cdh.data)
            print("\n")

        if not gh.compressed:
            print("Data starts at offset %s, length %s" % (
              self.reader.data_start_offset,
              self.reader.data_length
            ))
        else:
            mch = self.reader.main_compression_header
            cchs = self.reader.channel_compression_headers
            print("Main compression header starts at offset %s" % mch.offset)
            print("\n")
            for i, cch in enumerate(cchs):
                print(
                    "Channel compression header %s starts at offset %s" %
                    (i, cch.offset))
                print(cch.data)
                print(
                    "Compressed data starts at offset %s" %
                    cch.compressed_data_offset)
                print("\n")
        print("\n")
        print("Markers start at offset %s" % self.reader.marker_start_offset)


if __name__ == '__main__':
    main()



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# I'm just going to test one file, capture stdout
# Really I just want to make sure this thing basically runs so I don't
# push a totally broken executable in a release

from __future__ import absolute_import
from os import path

from bioread.runners import acq_info

DATA_PATH = path.join(path.dirname(path.abspath(__file__)), "data")

DATA_FILE = path.join(DATA_PATH, 'unicode', 'small-unicode-4.4.0.acq')


def test_acq_info_runs(capsys):
    acq_info.main([DATA_FILE])
    out, err = capsys.readouterr()
    assert len(out) > 0
    assert '2016-02-02' in out



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread
# Extended by Alexander Schlemmer.

from __future__ import with_statement, division
import struct
import zlib
from contextlib import contextmanager

import numpy as np

import bioread.file_revisions as rev
from bioread import headers as bh
from bioread.headers import GraphHeader, ChannelHeader, ChannelDTypeHeader
from bioread.headers import UnknownPaddingHeader
from bioread.headers import ForeignHeader, MainCompressionHeader
from bioread.headers import ChannelCompressionHeader
from bioread.headers import PostMarkerHeader, V2JournalHeader, V4JournalHeader
from bioread.headers import V4JournalLengthHeader
from bioread.biopac import Datafile, EventMarker

import logging
# Re-adding the handler on reload causes duplicate log messages.
logger = logging.getLogger("bioread")
logger.setLevel(logging.WARNING)
log_handler = logging.StreamHandler()
log_handler.setLevel(logging.DEBUG)
log_handler.setFormatter(logging.Formatter("%(message)s"))
if len(logger.handlers) == 0:  # Avoid duplicate messages on reload
    logger.addHandler(log_handler)


# This is how much interleaved uncompressed data we'll read at a time.
CHUNK_SIZE = 1024 * 256  # A suggestion, probably not a terrible one.

# How far past the foreign data header we're willing to go looking for the
# channel dtype headers
MAX_DTYPE_SCANS = 4096


class Reader(object):
    def __init__(self, acq_file=None):
        self.acq_file = acq_file
        self.encoding = None  # We're gonna guess from _set_order_and_version
        self.datafile = None
        # This must be set by _set_order_and_version
        self.byte_order_char = None
        self.file_revision = None
        self.samples_per_second = None
        self.graph_header = None
        self.channel_headers = []
        self.foreign_header = None
        self.channel_dtype_headers = []
        self.main_compression_header = None
        self.channel_compression_headers = []
        self.data_start_offset = None
        self.data_length = None
        self.marker_start_offset = None
        self.marker_header = None
        self.marker_item_headers = None
        self.event_markers = None

    @classmethod
    def read(cls,
             fo,
             channel_indexes=None,
             target_chunk_size=CHUNK_SIZE):
        """ Read a biopac file into memory.

        fo: The name of the file to read, or a file-like object
        channel_indexes: The numbers of the channels you want to read
        target_chunk_size: The amount of data to read in a chunk.

        returns: reader.Reader.
        """
        with open_or_yield(fo, 'rb') as io:
            reader = cls(io)
            reader._read_headers()
            reader._read_data(channel_indexes, target_chunk_size)
        return reader

    @classmethod
    def read_headers(cls, fo):
        """ Read only the headers -- no data -- of a biopac file.
        """
        with open_or_yield(fo, 'rb') as io:
            reader = cls(io)
            reader._read_headers()
        return reader

    def stream(self, channel_indexes=None, target_chunk_size=CHUNK_SIZE):
        """ Set up and retun an iterator for streaming data.
        """
        if self.datafile is None:
            self._read_headers()
        if self.is_compressed:
            raise TypeError('Streaming is not supported for compressed files')
        self.acq_file.seek(self.data_start_offset)
        return make_chunk_reader(
            self.acq_file,
            self.datafile.channels,
            channel_indexes,
            target_chunk_size)

    @property
    def is_compressed(self):
        return self.graph_header.compressed

    def _read_headers(self):
        logger.debug("I am in _read_headers")
        if self.byte_order_char is None:
            self.__set_order_and_version()

        self.graph_header = self.__single_header(0, GraphHeader)
        channel_count = self.graph_header.channel_count

        pad_start = self.graph_header.effective_len_bytes
        pad_headers = self.__multi_headers(
            self.graph_header.expected_padding_headers,
            pad_start,
            UnknownPaddingHeader)
        ch_start = pad_start + sum(
            [ph.effective_len_bytes for ph in pad_headers])
        pad_header = self.__single_header(ch_start, UnknownPaddingHeader)
        # if pad_header.effective_len_bytes == 40:
        #     ch_start = 0
        self.channel_headers = self.__multi_headers(channel_count,
                                                    ch_start, ChannelHeader)
        ch_len = self.channel_headers[0].effective_len_bytes

        for i, ch in enumerate(self.channel_headers):
            logger.debug("Channel header %s: %s" % (i, ch.data))

        fh_start = ch_start + len(self.channel_headers)*ch_len
        self.foreign_header = self.__single_header(fh_start, ForeignHeader)

        cdh_start = fh_start + self.foreign_header.effective_len_bytes
        self.channel_dtype_headers = self.__scan_for_dtype_headers(
            cdh_start, channel_count)
        if self.channel_dtype_headers is None:
            raise ValueError("Can't find valid channel data type headers")

        for i, cdt in enumerate(self.channel_dtype_headers):
            logger.debug("Channel %s: type_code: %s, offset: %s" % (
                i, cdt.type_code, cdt.offset
            ))

        logger.debug("Computed data start offset: %s" % self.data_start_offset)

        self.samples_per_second = 1000/self.graph_header.sample_time

        logger.debug("About to allocate a Datafile")
        self.datafile = Datafile(
            graph_header=self.graph_header,
            channel_headers=self.channel_headers,
            foreign_header=self.foreign_header,
            channel_dtype_headers=self.channel_dtype_headers,
            samples_per_second=self.samples_per_second)

        logger.debug("Allocated a datafile!")

        self.data_length = self.datafile.data_length
        logger.debug("Computed data length: %s" % self.data_length)

        # In compressed files, markers come before compressed data. But
        # data_length is 0 for compressed files.
        self.marker_start_offset = (self.data_start_offset + self.data_length)
        self._read_markers()
        try:
            self._read_journal()
        except struct.error:
            logger.info("No journal information found.")
        if self.is_compressed:
            self.__read_compression_headers()

    def __scan_for_dtype_headers(self, start_index, channel_count):
        # Sometimes the channel dtype headers don't seem to be right after the
        # foreign data header, and I can't find anything that directs me to the
        # proper location.
        # As a gross hack, we can scan forward until we find something
        # potentially valid.
        # Return a set of channel dtype headers when we find something valid,
        # self.data_start_offset will be set to the start of the data, and
        # and self.acq_file will be seek()ed to that location.
        logger.debug('Scanning for start of channel dtype headers')
        for i in range(MAX_DTYPE_SCANS):
            dtype_headers = self.__multi_headers(
                channel_count, start_index + i, ChannelDTypeHeader)
            if all([h.possibly_valid for h in dtype_headers]):
                logger.debug("Found at %s" % (start_index + i))
                self.data_start_offset = self.acq_file.tell()
                return dtype_headers
        logger.warn(
            "Couldn't find valid dtype headers, tried %s times" %
            MAX_DTYPE_SCANS
        )
        return None

    def __read_compression_headers(self):
        # We need to have read the markers and journal; this puts us
        # at the correct file offset.
        self.marker_start_offset = self.data_start_offset
        main_ch_start = self.acq_file.tell()
        self.main_compression_header = self.__single_header(
            main_ch_start, MainCompressionHeader)
        cch_start = (main_ch_start +
                     self.main_compression_header.effective_len_bytes)
        self.channel_compression_headers = self.__multi_headers(
            self.graph_header.channel_count, cch_start,
            ChannelCompressionHeader)

    def _read_journal(self):
        self.journal = None
        self.journal_header = None
        if self.file_revision <= rev.V_400B:
            self.__read_journal_v2()
        else:
            self.__read_journal_v4()
        self.datafile.journal_header = self.journal_header
        self.datafile.journal = self.journal

    def __read_journal_v2(self):
        self.post_marker_header = self.__single_header(
            self.acq_file.tell(), PostMarkerHeader)
        logger.debug("Reading journal starting at %s" % self.acq_file.tell())
        logger.debug(self.post_marker_header.rep_bytes)
        self.acq_file.seek(self.post_marker_header.rep_bytes, 1)
        logger.debug(self.acq_file.tell())
        self.journal_header = self.__single_header(
            self.acq_file.tell(), V2JournalHeader)
        self.journal = self.acq_file.read(
            self.journal_header.data['lJournalLen']).decode(
                self.encoding, errors='ignore').strip('\0')

    def __read_journal_v4(self):
        self.journal_length_header = self.__single_header(
            self.acq_file.tell(),
            V4JournalLengthHeader)
        journal_len = self.journal_length_header.journal_len
        self.journal = None
        jh = V4JournalHeader(
            self.file_revision, self.byte_order_char)
        # If journal_length_header.journal_len is small, we don't have a
        # journal to read.
        if (jh.effective_len_bytes <= journal_len):
            self.journal_header = self.__single_header(
                self.acq_file.tell(),
                V4JournalHeader)
            logger.debug("Reading {0} bytes of journal at {1}".format(
                self.journal_header.journal_len,
                self.acq_file.tell()))
            self.journal = self.acq_file.read(
                self.journal_header.journal_len).decode(
                    self.encoding, errors='ignore').strip('\0')
        # Either way, we should seek to this point.
        self.acq_file.seek(self.journal_length_header.data_end)

    def __single_header(self, start_offset, h_class):
        return self.__multi_headers(1, start_offset, h_class)[0]

    def __multi_headers(self, num, start_offset, h_class):
        headers = []
        last_h_len = 0  # This will be changed reading the channel headers
        h_offset = start_offset
        for i in range(num):
            h_offset += last_h_len
            logger.debug("Reading {0} at offset {1}".format(h_class, h_offset))
            h = h_class(self.file_revision,
                        self.byte_order_char,
                        encoding=self.encoding)
            h.unpack_from_file(self.acq_file, h_offset)
            logger.debug("Read %s bytes: %s" % (
                h.struct_dict.len_bytes, h.data))
            last_h_len = h.effective_len_bytes
            headers.append(h)
        return headers

    def _read_data(self, channel_indexes, target_chunk_size=CHUNK_SIZE):
        if self.is_compressed:
            self.__read_data_compressed(channel_indexes)
        else:
            self.__read_data_uncompressed(channel_indexes, target_chunk_size)

    def _read_markers(self):
        if self.marker_start_offset is None:
            self.read_headers()
        logger.debug("Reading markers starting at %s" %
            self.marker_start_offset)
        mh_class = bh.V2MarkerHeader
        mih_class = bh.V2MarkerItemHeader
        if self.file_revision >= rev.V_400B:
            mh_class = bh.V4MarkerHeader
            mih_class = bh.V4MarkerItemHeader
        self.marker_header = self.__single_header(
            self.marker_start_offset, mh_class)
        self.datafile.marker_header = self.marker_header
        self.__read_marker_items(mih_class)

    def __read_marker_items(self, marker_item_header_class):
        """
        self.acq_file must be seek()ed to the start of the first item header
        """
        event_markers = []
        marker_item_headers = []
        for i in range(self.marker_header.marker_count):
            mih = self.__single_header(
                self.acq_file.tell(), marker_item_header_class)
            marker_text_bytes = self.acq_file.read(mih.text_length)
            marker_text = marker_text_bytes.decode(
                self.encoding, errors='ignore').strip('\0')
            marker_item_headers.append(mih)
            marker_channel = self.datafile.channel_order_map.get(
                mih.channel_number)
            event_markers.append(EventMarker(
                time_index=(mih.sample_index * self.graph_header.sample_time) / 1000,
                sample_index=mih.sample_index,
                text=marker_text,
                channel_number=mih.channel_number,
                channel=marker_channel,
                date_created_ms=mih.date_created_ms,
                type_code=mih.type_code))
        self.marker_item_headers = marker_item_headers
        self.datafile.marker_item_headers = marker_item_headers
        self.datafile.event_markers = event_markers

    def __read_data_compressed(self, channel_indexes):
        # At least in post-4.0 files, the compressed data isn't interleaved at
        # all. It's stored in uniform compressed blocks -- this probably
        # compresses far better than interleaved data.
        # Strangely, the compressed data seems to always be little-endian.
        if channel_indexes is None:
            channel_indexes = np.arange(len(self.datafile.channels))

        for i in channel_indexes:
            cch = self.channel_compression_headers[i]
            channel = self.datafile.channels[i]
            # Data seems to always be little-endian
            dt = channel.dtype.newbyteorder("<")
            self.acq_file.seek(cch.compressed_data_offset)
            comp_data = self.acq_file.read(cch.compressed_data_len)
            decomp_data = zlib.decompress(comp_data)
            channel.raw_data = np.frombuffer(decomp_data, dtype=dt)

    def __read_data_uncompressed(self, channel_indexes, target_chunk_size):
        self.acq_file.seek(self.data_start_offset)
        # This will fill self.datafile.channels with data.
        read_uncompressed(
            self.acq_file,
            self.datafile.channels,
            channel_indexes,
            target_chunk_size)

    def __set_order_and_version(self):
        # Try unpacking the version string in both a bid and little-endian
        # fashion. Version string should be a small, positive integer.
        self.acq_file.seek(0)
        # No byte order flag -- we're gonna figure it out.
        gh = GraphHeader(rev.V_ALL, '')
        ver_fmt_str = gh.format_string
        ver_len = struct.calcsize('<'+ver_fmt_str)
        ver_data = self.acq_file.read(ver_len)

        byte_order_chars = ['<', '>']
        # Try both ways.
        byte_order_versions = [
            (struct.unpack(boc+ver_fmt_str, ver_data)[1], boc)
            for boc in byte_order_chars
        ]

        # Limit to positive numbers, choose smallest.
        byte_order_versions = sorted([
            bp for bp in byte_order_versions if bp[0] > 0])
        bp = byte_order_versions[0]

        self.byte_order_char = bp[1]
        self.file_revision = bp[0]
        # Guess at file encoding -- I think that everything before acq4 is
        # in latin1 and everything newer is utf-8
        logger.debug("File revision: %s" % self.file_revision)
        logger.debug("Byte order: %s" % self.byte_order_char)
        if self.file_revision < rev.V_400B:
            self.encoding = 'latin1'
        else:
            self.encoding = 'utf-8'

    def __repr__(self):
        return "Reader('{0}')".format(self.acq_file)


@contextmanager
def open_or_yield(thing, mode):
    """ If 'thing' is a string, open it and yield it. Otherwise, yield it.

    This lets you use a filename, open file, other IO object. If 'thing' was
    a filename, the file is guaranteed to be closed after yielding.
    """
    if isinstance(thing, str):
        with open(thing, mode) as f:
            yield(f)
    else:
        yield(thing)


class ChunkBuffer(object):
    def __init__(self, channel):
        self.channel = channel
        self.buffer = None
        self.channel_slice = slice(0, 0)


def read_uncompressed(
        f,
        channels,
        channel_indexes=None,
        target_chunk_size=CHUNK_SIZE):
    """
    Read the uncompressed data.

    This function will read the data from an open IO object f (which must be
    seek()ed to the start of the data) into the raw_data attribute of
    channels.

    channel_indexes is a list of indexes of the channels we want to read
    (if None, read all the channels). Other channels' raw_data will be set
    to None.

    target_chunk_size gives a general idea of how much data the program should
    read into memory at a time. You can probably always leave this as at its
    default.

    This function returns nothing; it modifies channels in-place.

    Uncompressed data are stored in .acq files in an interleaved format --
    as the data streams off the amps, it's stored directly. So, with three
    channels, your data might look like (spaces added for clarity):

    012 012 012 012 012 ...

    Each channel can also have a frequency divider, which tells us this
    channel is recorded every nth occurence of the file's base sampling rate.

    If our three channels have frequency dividers [1, 4, 2], the data pattern
    would look like (again, with spaces between repetitions):
    0120020 0120020 0120020 ...

    """
    if channel_indexes is None:
        channel_indexes = np.arange(len(channels))

    for i in channel_indexes:
        channels[i]._allocate_raw_data()

    chunker = make_chunk_reader(
        f, channels, channel_indexes, target_chunk_size)
    for chunk_buffers in chunker:
        for i in channel_indexes:
            ch = channels[i]
            buf = chunk_buffers[i]
            logger.debug('Storing {0} samples to {1} of channel {2}'.format(
                len(buf.buffer), buf.channel_slice, i))
            ch.raw_data[buf.channel_slice] = buf.buffer[:]


def make_chunk_reader(
        f,
        channels,
        channel_indexes=None,
        target_chunk_size=CHUNK_SIZE):

    if channel_indexes is None:
        channel_indexes = np.arange(len(channels))

    byte_pattern = chunk_byte_pattern(channels, target_chunk_size)
    logger.debug('Using chunk size: {0} bytes'.format(len(byte_pattern)))
    buffers = [ChunkBuffer(c) for c in channels]
    return read_chunks(f, buffers, byte_pattern, channel_indexes)


def read_chunks(f, buffers, byte_pattern, channel_indexes):
    """
    Read data in chunks from f. For each chunk, yield a list of buffers with
    information on how much of the buffer is filled and where the data should
    go in the target array.
    """
    channel_bytes_remaining = np.array(
        [b.channel.data_length for b in buffers])
    chunk_number = 0
    while np.sum(channel_bytes_remaining) > 0:
        pat = chunk_pattern(byte_pattern, channel_bytes_remaining)
        chunk_bytes = len(pat)
        logger.debug('Chunk {0}: {1} bytes at {2}'.format(
            chunk_number, chunk_bytes, f.tell()))
        chunk_data = np.frombuffer(
            f.read(chunk_bytes), dtype="b", count=chunk_bytes)
        update_buffers_with_data(
            chunk_data, buffers, pat, channel_indexes)

        yield buffers
        channel_bytes_remaining -= np.bincount(
            pat, minlength=len(channel_bytes_remaining))
        logger.debug('Channel bytes remaining: {0}'.format(
            channel_bytes_remaining))
        chunk_number += 1


def chunk_pattern(byte_pattern, channel_bytes_remaining):
    """ Trim a byte pattern depending on how many bytes remain in each channel.

    For some reason, the data at the end of the file doesn't work like you'd
    expect. You can, for example, be missing an expected sample in a slow-
    sampling channel.

    The solution is to use the number of bytes in a channel to determine the
    actual layout of the chunk.
    """
    # This is the normal case, we don't need to do anything.
    if np.all(np.bincount(byte_pattern) <= channel_bytes_remaining):
        return byte_pattern
    # For each channel, compute a set of indexes where we expect data.
    channel_byte_indexes = [
        np.where(byte_pattern == i)[0][0:rem]
        for i, rem in enumerate(channel_bytes_remaining)
    ]
    all_byte_indexes = np.concatenate(channel_byte_indexes)
    pattern_mask = np.zeros(len(byte_pattern), dtype=bool)
    pattern_mask[all_byte_indexes] = True
    return byte_pattern[pattern_mask]


def update_buffers_with_data(data, buffers, byte_pattern, channel_indexes):
    """
    Updates buffers with information from data. Returns nothing, modifies
    buffers in-place.
    """
    trimmed_pattern = byte_pattern[0:len(data)]
    for i in channel_indexes:
        buf = buffers[i]
        buf.buffer = data[trimmed_pattern == i]
        buf.buffer.dtype = buf.channel.dtype
        old_slice = buf.channel_slice
        buf.channel_slice = slice(
            old_slice.stop, old_slice.stop + len(buf.buffer))


def chunk_byte_pattern(channels, target_chunk_size):
    """ Compute a byte layout for a chunk of data.

    This pattern is the main thing we actually need -- from it, we can know
    how to make individual buffers and how much data to read.

    The actual chunk size will always be a multiple of the byte pattern
    length, and will generally be very close to target_chunk_size. Usually, it
    will be larger.
    """
    divs = np.array([c.frequency_divider for c in channels])
    sizes = np.array([c.sample_size for c in channels])
    spat = sample_pattern(divs)
    byte_counts = sizes[spat]  # Returns array the length of spat
    bpat = spat.repeat(byte_counts)
    reps = chunk_pattern_reps(target_chunk_size, len(bpat))
    return np.tile(bpat, reps)


def sample_pattern(frequency_dividers):
    """ Compute the pattern of samples in a file's uncompressed data stream.

    The basic algorithm:
    * Take the least common multiple of the frequency dividers. This is the
      "base" of the pattern length -- the most times a channel could appear in
      the pattern.
    * Make a [base_len x num_channels] dimension matrix, counting from 0 to
      pattern_len in each row -- call this "pattern_slots"
    * Make a pattern_mask -- a boolean mask where each channel slots modulo
      frequency_divider == 0
    * The pattern, then, are the pattern_slots where pattern_mask is true

    Note that this is not quite the byte pattern -- these samples can either
    be int16 or float64.
    """
    dividers = np.array(frequency_dividers)
    channel_count = len(dividers)
    base_len = least_common_multiple(*dividers)
    pattern_slots = np.arange(
        base_len).repeat(
        channel_count).reshape(
        (base_len, channel_count))
    pattern_mask = ((pattern_slots % dividers) == 0)
    channel_slots = np.tile(np.arange(channel_count), (base_len, 1))
    return channel_slots[pattern_mask]


def chunk_pattern_reps(target_chunk_size, pattern_byte_length):
    """
    The number of times we'll actually repeat the pattern in a chunk.
    Must always be at least 1.
    """
    return max(1, target_chunk_size // pattern_byte_length)


def least_common_multiple(*ar):
    """ Compute least common multiple of n numbers.

    Adapted from:
    http://stackoverflow.com/questions/147515/least-common-multiple-for-3-or-more-numbers

    Used in computing the repeating pattern of multichannel data that's
    sampled at different rates in each channel.
    """

    if len(ar) > 2:
        return least_common_multiple(ar[0], least_common_multiple(*ar[1:]))
    elif len(ar) == 2:
        return (ar[0] * ar[1]) // greatest_common_denominator(ar[0], ar[1])
    else:
        return ar[0]


def greatest_common_denominator(a, b):
    """ Iterative method to compute greatest common denominator. """
    while not b == 0:
        a, b = b, a % b
    return a



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# flake8: noqa: E203, E231, E741, E405, E403

# If I were to do this over again, I'd probably design it differently, without
# the somewhat weird versioning structure, as well as using ctypes instead
# of the weird StructDict thing we have.
# But, you know what? It's not worth changing this now, unless there turns out
# to be a compelling reason in the future.
# But if you're ever looking to do something similar, probably don't use this as
# a template of good code.


from bioread.struct_dict import StructDict
from bioread.file_revisions import *


class Header(object):
    """
    Represents a binary header and its unpacking
    """

    def __init__(self, struct_dict, encoding='utf-8'):
        self.struct_dict = struct_dict
        self.encoding = encoding
        self.offset = None
        self.unpacked = {}

    def unpack_from_str(self, str_data):
        self.raw_data = str_data
        self.__unpack_data()

    def unpack_from_file(self, data_file, offset):
        self.offset = offset
        data_file.seek(offset)
        self.raw_data = data_file.read(self.struct_dict.len_bytes)
        self.__unpack_data()

    @property
    def effective_len_bytes(self):
        """
        This will be overridden frequently -- it's used in navigating files.
        """
        return self.struct_dict.len_bytes

    @property
    def format_string(self):
        return self.struct_dict.format_string

    def __unpack_data(self):
        self.data = self.struct_dict.unpack(self.raw_data)

    def __getitem__(self, key):
        return self.data[key]

    def __str__(self):
        return str(self.data)


class VersionedHeaderStructure(object):

    def __init__(self, *structure_elements):
        self.structure_elements = structure_elements

    def elements_for(self, version):
        return [se for se in self.structure_elements if se[2] <= version]


class BiopacHeader(Header):
    """
    A simple superclass for GraphHeader, ChannelHeader, and friends.
    """

    def __init__(self, header_structure, file_revision, byte_order_char,
                 **kwargs):
        self.file_revision = file_revision
        self.byte_order_char = byte_order_char
        self.header_structure = header_structure
        sd = StructDict(byte_order_char,
                        header_structure.elements_for(file_revision))
        super().__init__(sd, **kwargs)


class GraphHeader(BiopacHeader):
    """
    The main Graph Header for an AcqKnowledge file. The documentation is much
    more complete for older files (revision 45 and below); I've looked to find
    the reliable (and essential) structure information in later files.

    The fields we really need to decode the data:
    Length
    Number of channels
    Sample time
    Data compressed?
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        self.file_revision = file_revision
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def effective_len_bytes(self):
        l = self.data['lExtItemHeaderLen']
        # We're going to skip an extra 40 bytes. This is kind of an ugly hack;
        # there seems to be a whole other 40-byte header in there.
        # if self.file_revision >= V_430:
        #     l += 40
        return l

    @property
    def channel_count(self):
        return self.data['nChannels']

    @property
    def sample_time(self):
        return self.data['dSampleTime']

    @property
    def compressed(self):
        return self.data.get('bCompressed') == 1

    @property
    def data_format(self):
        fmt = 'uncompressed'
        if self.compressed:
            fmt = 'compressed'
        return fmt

    @property
    def __version_bin(self):
        bin = 'Unknown'
        if self.file_revision < V_400B:
            bin = 'PRE_4'
        else:
            bin = 'POST_4'
        return bin

    @property
    def expected_padding_headers(self):
        if self.file_revision >= V_430:
            return self.data['hExpectedPaddings']
        return 0

    @property
    def __h_elt_versions(self):
        return {
            'PRE_4' : VersionedHeaderStructure(
                ('nItemHeaderLen'           ,'h'    ,V_ALL),
                ('lVersion'                 ,'l'    ,V_ALL),
                ('lExtItemHeaderLen'        ,'l'    ,V_20a),
                ('nChannels'                ,'h'    ,V_20a),
                ('nHorizAxisType'           ,'h'    ,V_20a),
                ('nCurChannel'              ,'h'    ,V_20a),
                ('dSampleTime'              ,'d'    ,V_20a),
                ('dTimeOffset'              ,'d'    ,V_20a),
                ('dTimeScale'               ,'d'    ,V_20a),
                ('dTimeCursor1'             ,'d'    ,V_20a),
                ('dTimeCursor2'             ,'d'    ,V_20a),
                ('rcWindow'                 ,'4h'   ,V_20a),
                ('nMeasurement'             ,'6h'   ,V_20a),
                ('fHilite'                  ,'h'    ,V_20a),
                ('dFirstTimeOffset'         ,'d'    ,V_20a),
                ('nRescale'                 ,'h'    ,V_20a),
                ('szHorizUnits1'            ,'40s'  ,V_20a),
                ('szHorizUnits2'            ,'10s'  ,V_20a),
                ('nInMemory'                ,'h'    ,V_20a),
                ('fGrid'                    ,'h'    ,V_20a),
                ('fMarkers'                 ,'h'    ,V_20a),
                ('nPlotDraft'               ,'h'    ,V_20a),
                ('nDispMode'                ,'h'    ,V_20a),
                ('rRReserved'               ,'h'    ,V_20a),
                ('BShowToolBar'             ,'h'    ,V_30r),
                ('BShowChannelButtons'      ,'h'    ,V_30r),
                ('BShowMeasurements'        ,'h'    ,V_30r),
                ('BShowMarkers'             ,'h'    ,V_30r),
                ('BShowJournal'             ,'h'    ,V_30r),
                ('CurXChannel'              ,'h'    ,V_30r),
                ('MmtPrecision'             ,'h'    ,V_30r),
                ('NMeasurementRows'         ,'h'    ,V_303),
                ('mmt40'                    ,'40h'  ,V_303),
                ('mmtChan40'                ,'40h'  ,V_303),
                ('MmtCalcOpnd1'             ,'40h'  ,V_35x),
                ('MmtCalcOpnd2'             ,'40h'  ,V_35x),
                ('MmtCalcOp'                ,'40h'  ,V_35x),
                ('MmtCalcConstant'          ,'40d'  ,V_35x),
                ('bNewGridWithMinor'        ,'l'    ,V_370),
                ('colorMajorGrid'           ,'4B'   ,V_370),
                ('colorMinorGrid'           ,'4B'   ,V_370),
                ('wMajorGridStyle'          ,'h'    ,V_370),
                ('wMinorGridStyle'          ,'h'    ,V_370),
                ('wMajorGridWidth'          ,'h'    ,V_370),
                ('wMinorGridWidth'          ,'h'    ,V_370),
                ('bFixedUnitsDiv'           ,'l'    ,V_370),
                ('bMid_Range_Show'          ,'l'    ,V_370),
                ('dStart_Middle_Point'      ,'d'    ,V_370),
                ('dOffset_Point'            ,'60d'  ,V_370),
                ('hGrid'                    ,'d'    ,V_370),
                ('vGrid'                    ,'60d'  ,V_370),
                ('bEnableWaveTools'         ,'l'    ,V_370),
                ('hozizPrecision'           ,'h'    ,V_373),
                ('Reserved'                 ,'20b'  ,V_381),
                ('bOverlapMode'             ,'l'    ,V_381),
                ('bShowHardware'            ,'l'    ,V_381),
                ('bXAutoPlot'               ,'l'    ,V_381),
                ('bXAutoScroll'             ,'l'    ,V_381),
                ('bStartButtonVisible'      ,'l'    ,V_381),
                ('bCompressed'              ,'l'    ,V_381),
                ('bAlwaysStartButtonVisible','l'    ,V_381),
                ('pathVideo'                ,'260s' ,V_382),
                ('optSyncDelay'             ,'l'    ,V_382),
                ('syncDelay'                ,'d'    ,V_382),
                ('bHRP_PasteMeasurements'   ,'l'    ,V_382),
                ('graphType'                ,'l'    ,V_390),
                ('mmtCalcExpr'              ,'10240s',V_390),
                ('mmtMomentOrder'           ,'40l'  ,V_390),
                ('mmtTimeDelay'             ,'40l'  ,V_390),
                ('mmtEmbedDim'              ,'40l'  ,V_390),
                ('mmtMIDelay'               ,'40l'  ,V_390),
            ),
            'POST_4' : VersionedHeaderStructure(
                ('nItemHeaderLen'           ,'h'    ,V_ALL),
                ('lVersion'                 ,'l'    ,V_ALL),
                ('lExtItemHeaderLen'        ,'l'    ,V_20a),
                ('nChannels'                ,'h'    ,V_20a),
                ('nHorizAxisType'           ,'h'    ,V_20a),
                ('nCurChannel'              ,'h'    ,V_20a),
                ('dSampleTime'              ,'d'    ,V_20a),
                ('dTimeOffset'              ,'d'    ,V_20a),
                ('dTimeScale'               ,'d'    ,V_20a),
                ('dTimeCursor1'             ,'d'    ,V_20a),
                ('dTimeCursor2'             ,'d'    ,V_20a),
                ('rcWindow'                 ,'4h'   ,V_20a),
                ('nMeasurement'             ,'6h'   ,V_20a),
                ('fHilite'                  ,'h'    ,V_20a),
                ('dFirstTimeOffset'         ,'d'    ,V_20a),
                ('nRescale'                 ,'h'    ,V_20a),
                ('szHorizUnits1'            ,'40s'  ,V_20a),
                ('szHorizUnits2'            ,'10s'  ,V_20a),
                ('nInMemory'                ,'h'    ,V_20a),
                ('fGrid'                    ,'h'    ,V_20a),
                ('fMarkers'                 ,'h'    ,V_20a),
                ('nPlotDraft'               ,'h'    ,V_20a),
                ('nDispMode'                ,'h'    ,V_20a),
                ('rRReserved'               ,'h'    ,V_20a),
                ('Unknown'                  ,'822B' ,V_400B),
                ('bCompressed'              ,'l'    ,V_400B),
                ('Unknown2'                 ,'1422B',V_400B),
                ('hExpectedPaddings'        ,'h'    ,V_430)
            )}

    @property
    def __h_elts(self):
        return self.__h_elt_versions[self.__version_bin]


class UnknownPaddingHeader(BiopacHeader):
    """
    I don't know what this is for, but it's 40-bytes long and right before some
    modern files
    """
    def __init__(self, file_revision, byte_order_char, **kwargs):
        self.file_revision = file_revision
        super().__init__(
            self.__h_elts, file_revision, byte_order_char, **kwargs)

    @property
    def effective_len_bytes(self):
        return self.data['lChannelLen']

    @property
    def __h_elts(self):
        elts = VersionedHeaderStructure(
            ('lChannelLen', 'l', V_ALL),
            ('Uknown', '36B', V_ALL)
        )
        return elts


class ChannelHeader(BiopacHeader):
    """
    The main Channel Header for an AcqKnowledge file. Note that this is known
    to be wrong for more modern files -- but for the purposes of this
    reader, we don't care. Enough fields are right that we can still find
    our way around.
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        self.file_revision = file_revision
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def __version_bin(self):
        bin = 'Unknown'
        if self.file_revision < V_400B:
            bin = 'PRE_4'
        else:
            bin = 'POST_4'
        return bin

    @property
    def effective_len_bytes(self):
        return self.data['lChanHeaderLen']

    @property
    def frequency_divider(self):
        """
        The number you divide the graph's samples_per_second by to get this
        channel's samples_per_second. If not defined (as in old file
        versions), this should be 1.
        """
        return self.data.get('nVarSampleDivider') or 1

    @property
    def raw_scale(self):
        return self.data['dAmplScale']

    @property
    def raw_offset(self):
        return self.data['dAmplOffset']

    @property
    def units(self):
        return self.data['szUnitsText'].decode(
            self.encoding, errors='ignore').strip('\0')

    @property
    def name(self):
        return self.data['szCommentText'].decode(
            self.encoding, errors='ignore').strip('\0')

    @property
    def point_count(self):
        return self.data['lBufLength']

    @property
    def order_num(self):
        return self.data['nChanOrder']

    @property
    def __h_elts(self):
        return self.__h_elt_versions[self.__version_bin]

    @property
    def __h_elt_versions(self):
        return {
            'PRE_4' : VersionedHeaderStructure(
                ('lChanHeaderLen'           ,'l'    ,V_20a),
                ('nNum'                     ,'h'    ,V_20a),
                ('szCommentText'            ,'40s'  ,V_20a),
                ('rgbColor'                 ,'4B'   ,V_20a),
                ('nDispChan'                ,'h'    ,V_20a),
                ('dVoltOffset'              ,'d'    ,V_20a),
                ('dVoltScale'               ,'d'    ,V_20a),
                ('szUnitsText'              ,'20s'  ,V_20a),
                ('lBufLength'               ,'l'    ,V_20a),
                ('dAmplScale'               ,'d'    ,V_20a),
                ('dAmplOffset'              ,'d'    ,V_20a),
                ('nChanOrder'               ,'h'    ,V_20a),
                ('nDispSize'                ,'h'    ,V_20a),
                ('plotMode'                 ,'h'    ,V_30r),
                ('vMid'                     ,'d'    ,V_30r),
                ('szDescription'            ,'128s' ,V_370),
                ('nVarSampleDivider'        ,'h'    ,V_370),
                ('vertPrecision'            ,'h'    ,V_373),
                ('activeSegmentColor'       ,'4b'   ,V_382),
                ('activeSegmentStyle'       ,'l'    ,V_382),
            ),
            'POST_4' : VersionedHeaderStructure(
                ('lChanHeaderLen'           ,'l'    ,V_20a),
                ('nNum'                     ,'h'    ,V_20a),
                ('szCommentText'            ,'40s'  ,V_20a),
                ('notColor'                 ,'4B'   ,V_20a),
                ('nDispChan'                ,'h'    ,V_20a),
                ('dVoltOffset'              ,'d'    ,V_20a),
                ('dVoltScale'               ,'d'    ,V_20a),
                ('szUnitsText'              ,'20s'  ,V_20a),
                ('lBufLength'               ,'l'    ,V_20a),
                ('dAmplScale'               ,'d'    ,V_20a),
                ('dAmplOffset'              ,'d'    ,V_20a),
                ('nChanOrder'               ,'h'    ,V_20a),
                ('nDispSize'                ,'h'    ,V_20a),
                ('unknown'                  ,'40s'  ,V_400B),
                ('nVarSampleDivider'        ,'h'    ,V_400B),
            )}


class ForeignHeader(BiopacHeader):
    """
    The Foreign Data header for AcqKnowledge files. I admit that I don't
    know what goes in this section.
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        self.file_revision = file_revision
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def __version_bin(self):
        bin = 'Unknown'
        if self.file_revision <= V_390:
            bin = "PRE_4"
        elif self.file_revision >= V_400B:
            bin = "POST_4"
        return bin

    @property
    def effective_len_bytes(self):
        if self.__version_bin == "PRE_4":
            return self.data['nLength']
        return self.data['lLength']

    @property
    def __h_elts(self):
        return self.__h_elt_versions[self.__version_bin]

    @property
    def __h_elt_versions(self):
        return {
            "PRE_4" : VersionedHeaderStructure(
                ('nLength'                  ,'h'    ,V_20a),
                ('nType'                    ,'h'    ,V_20a),
            ),
            "POST_4" : VersionedHeaderStructure(
                ('lLength'                  ,'l'    ,V_400B),
            )}


class ChannelDTypeHeader(BiopacHeader):
    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def type_code(self):
        return self.data['nType']

    CODE_MAP = {
        0: 'f8',
        1: 'f8',
        2: 'i2'
    }

    @property
    def possibly_valid(self):
        # TODO: When and if we someday figure out the post-foreign-data header
        # mystery, remove this. Until then, we'll use this function to guess at
        # the location of the channel dtype header set.
        dtype_code = self.CODE_MAP.get(self.type_code, None)
        if dtype_code is None:
            return False
        type_size = int(dtype_code[-1])
        return type_size == self.sample_size

    @property
    def numpy_dtype(self):
        """
        A valid numpy dtype string: byte order and data type.

        Examples: '<i2', '>f8'
        """
        return self.byte_order_char + self.CODE_MAP[self.type_code]

    @property
    def sample_size(self):
        return self.data['nSize']

    @property
    def __h_elts(self):
        # This lets the standard effective_len_bytes work fine, I think.
        return VersionedHeaderStructure(
        ('nSize'                    ,'h'    ,V_20a),
        ('nType'                    ,'h'    ,V_20a),
        )


class PostMarkerHeader(BiopacHeader):
    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
            ('hUnknown1', 'h', V_20a),
            ('hUnknown2', 'h', V_20a),
            ('lReps', 'l', V_20a),
            ('Unknown3', '80B', V_20a)
        )

    @property
    def effective_len_bytes(self):
        return self.struct_dict.len_bytes + self.rep_bytes

    @property
    def rep_bytes(self):
        # We seem to always need to read at least one of these...
        # I have no idea what these things are for. From what I've seen, the
        # file repeats the same data over and over, with a leading counter
        # byte. The last four bytes in the structure are 44 33 22 11.
        # The journal header follows immediately afterwards.
        reps = self.data['lReps']
        if reps < 1:
            reps = 1
        return 28 * reps


class V2JournalHeader(BiopacHeader):
    """
    Version 2-3 journal headers are trivial -- there's two bytes of
    "I don't know what this is" followed by a long containing the length of
    the journal text.
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
            ('hUnknown', 'h', V_20a),
            ('lJournalLen', 'l', V_20a)
        )


class V4JournalLengthHeader(BiopacHeader):
    """
    In the case where there's no journal data, there's no full journal header.
    Instead, we just have a single long that tells us how much journal stuff
    (data + header) there is. Basically, if this value is less than the
    length of the V4JournalHeader, don't even try to read that header or
    journal data.

    The next stuff (if there is any) will be at self.offset + lJournalDataLen
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
            ('lJournalDataLen', 'l', V_400B)
        )

    @property
    def journal_len(self):
        return self.data['lJournalDataLen']

    @property
    def data_end(self):
        return self.offset + self.journal_len



class V4JournalHeader(BiopacHeader):
    """
    In Version 4.1 and less, the journal is stored as plain text. From 4.2,
    it's stored as HTML. The start of the header tells the length of the
    entire journal section -- journal text and some preamble; the compression
    headers (if compressed) follow at self.offset + lFullLength.

    The length of the actual journal text is contained 266 bytes into the
    preamble for 4.1 files; in 4.2 files it's a long immediately before the
    journal data.

    The journal data starts at 560 bytes after the start of this header in
    4.1 files; in 4.2 it's at 594 bytes, and 4.4 it's at 598 bytes.
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
            ('bUnknown1', '262b', V_400B),
            ('lEarlyJournalLen', 'l', V_400B),
            ('bUnknown2', '290b', V_400B),
            ('bUnknown3', '26b', V_420),
            ('bUnknown4', '4b', V_440),
            ('lLateJournalLenMinusOne', 'l', V_420),
            ('lLateJournalLen', 'l', V_420)
        )

    @property
    def journal_len(self):
        if self.file_revision < V_420:
            return self.data['lEarlyJournalLen']
        return self.data['lLateJournalLen']


class MainCompressionHeader(BiopacHeader):
    # In compressed files, the markers are stored where the data would be in
    # uncompressed files. There's also some padding, and I don't know
    # what's in it.
    def __init__(self, file_revision, byte_order_char, **kwargs):
        self.file_revision = file_revision
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def effective_len_bytes(self):
        return self.__effective_len_byte_versions[self.__version_bin]()

    @property
    def __effective_len_byte_versions(self):
        # Determined through experimentation, may not be correct for some
        # revisions. Or files, for that matter. We'll see.
        return {
            'PRE_4': lambda: (
                self.struct_dict.len_bytes + self.data['lTextLen']),
            'POST_4': lambda: (  # This is incorrect for data with journals
                self.struct_dict.len_bytes +
                self.data['lStrLen1'] +
                self.data['lStrLen2']
                )
        }

    @property
    def __version_bin(self):
        bin = "Unknown"
        if self.file_revision <= V_400B:
            bin = "PRE_4"
        else:
            bin = "POST_4"
        return bin

    @property
    def __h_elts(self):
        return self.__h_elts_versions[self.__version_bin]

    @property
    def __h_elts_versions(self):
        return {
            'PRE_4': VersionedHeaderStructure(
                ('Unknown', '34B', V_20a),
                ('lTextLen', 'l', V_20a)
            ),
            'POST_4': VersionedHeaderStructure(
                ('Unknown1', '24B', V_400B),  # Should probably be 24.
                ('lStrLen1', 'l', V_400B),
                ('lStrLen2', 'l', V_400B),
                ('Unknown2', '20B', V_400B),
                ('Unknown3', '6B', V_420),
            )
        }


class ChannelCompressionHeader(BiopacHeader):
    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def effective_len_bytes(self):
        """
        Return the length of the header UP TO THE NEXT HEADER, skipping the
        compressed data. Use header_only_len_bytes for only the header length.
        """
        return self.header_only_len_bytes + self.data['lCompressedLen']

    @property
    def header_only_len_bytes(self):
        """
        A truly variable-length header. Oddly, the channel description and
        units are included in the header. This class doesn't have a way to
        get those out, but we'll just use them from the earlier channel
        header.
        Immediately after this header, the compressed data starts -- it will
        be after the units text, and starts with 'x'.
        """
        return (
            self.struct_dict.len_bytes + self.data['lChannelLabelLen'] +
            self.data['lUnitLabelLen'])

    @property
    def compressed_data_offset(self):
        """
        The offset to the compressed data.
        Note that this won't be valid until self#unpack_from_file() is run.
        """
        return self.offset + self.header_only_len_bytes

    @property
    def compressed_data_len(self):
        return self.data['lCompressedLen']

    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
        ('Unknown'                  ,'44B'  ,V_381),
        ('lChannelLabelLen'         ,'l'    ,V_381),
        ('lUnitLabelLen'            ,'l'    ,V_381),
        ('lUncompressedLen'         ,'l'    ,V_381),
        ('lCompressedLen'           ,'l'    ,V_381),
        )


class V2MarkerHeader(BiopacHeader):
    """
    Marker structure for files in Version 3, very likely down to version 2.
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    # NOTE: lLength does NOT include this header length -- only the length
    # of all the marker items. This is different than in the v4 header.
    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
        ('lLength'              ,'l'    ,V_20a),
        ('lMarkers'             ,'l'    ,V_20a),
        )

    @property
    def marker_count(self):
        return self.data['lMarkers']


class V4MarkerHeader(BiopacHeader):
    """
    Marker structure for files from Version 4 onwards
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    # NOTE: lLength INCLUDES this header length -- the markers end at
    # marker_start_offset + lLength. This is different than in the v2 header.
    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
        ('lLength'              ,'l'    ,V_400B),
        ('lMarkersExtra'        ,'l'    ,V_400B),
        ('lMarkers'             ,'l'    ,V_400B),
        ('Unknown'              ,'6B'   ,V_400B),
        ('szDefl'               ,'5s'   ,V_400B),
        ('Unknown2'             ,'h'    ,V_400B),
        ('Unknown3'             ,'8B'   ,V_42x),
        ('Unknown4'             ,'8B'   ,V_440)
        )

    # I'm not quite sure about these two marker count headers; they seem to
    # both be wrong.
    @property
    def marker_count(self):
        return self.data['lMarkersExtra'] - 1


class V2MarkerItemHeader(BiopacHeader):
    """
    Marker Items for files in Version 3, very likely down to version 2.
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
        ('lSample'              ,'l'    ,V_20a),
        ('fSelected'            ,'h'    ,V_35x),
        ('fTextLocked'          ,'h'    ,V_20a),
        ('fPositionLocked'      ,'h'    ,V_20a),
        ('nTextLength'          ,'h'    ,V_20a),
        )

    # Note: The spec says nTextLength includes the trailing null, but it
    # seems to not...?
    # It seems that it does include it in V_303 so haha
    @property
    def text_length(self):
        if self.file_revision < V_35x:
            return self.data['nTextLength']
        else:
            return self.data['nTextLength'] + 1

    @property
    def sample_index(self):
        return self.data['lSample']

    @property
    def channel_number(self):
        """ None means it's a global marker """
        return None

    @property
    def date_created_ms(self):
        """ markers don't get "Date created" until v. 4.4.0 """
        return None

    @property
    def type_code(self):
        """ These markers don't get type_codes, but it's OK """
        return None



class V4MarkerItemHeader(BiopacHeader):
    """
    Marker Items for files in Version 4 ownards.
    """

    def __init__(self, file_revision, byte_order_char, **kwargs):
        super().__init__(self.__h_elts, file_revision, byte_order_char,
                         **kwargs)

    @property
    def __h_elts(self):
        return VersionedHeaderStructure(
        ('lSample'              ,'l'    ,V_400B),
        ('Unknown'              ,'4B'   ,V_400B),
        ('nChannel'             ,'h'    ,V_400B),
        ('sMarkerStyle'         ,'4s'   ,V_400B),
        ('llDateCreated'        ,'Q'    ,V_440),
        ('Unknown3'             ,'8B'   ,V_42x),
        ('nTextLength'          ,'h'    ,V_400B),
        )

    # Unlike in older versions, nTextLength does include the trailing null.
    @property
    def text_length(self):
        return self.data['nTextLength']

    @property
    def sample_index(self):
        return self.data['lSample']

    @property
    def channel_number(self):
        """ None means it's a global marker """
        chan = self.data['nChannel']
        if chan == -1:
            chan = None
        return chan

    @property
    def date_created_ms(self):
        """ Date when marker was created (in ms since 1970-01-01) """
        if self.file_revision < V_440:
            return None
        else:
            return self.data['llDateCreated']

    @property
    def type_code(self):
        return self.data['sMarkerStyle'].decode(self.encoding, errors='ignore')



# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

import numpy as np


class MatlabWriter(object):

    def __init__(
            self,
            data=None,
            filename=None,
            compress=False,
            oned_as='row',
            data_only=False):
        self.data = data
        self.filename = filename
        self.compress = compress
        self.oned_as = oned_as
        self.data_only = data_only

    @classmethod
    def write_file(
            cls, data, filename, compress=False, oned_as='row', data_only=False):
        writer = cls(data, filename, compress, oned_as, data_only)
        writer.write()

    def write(self):
        from scipy.io import savemat
        d = self.__build_dict(self.data)
        savemat(
            self.filename, d,
            format='5', do_compression=self.compress,
            oned_as=self.oned_as)

    def __build_dict(self, data):
        d = {}
        d['samples_per_second'] = data.samples_per_second
        nc = len(data.channels)
        channels = np.zeros(nc, dtype='O')
        channel_headers = np.zeros(nc, dtype='O')
        channel_dtype_headers = np.zeros(nc, dtype='O')
        for i in range(nc):
            c = data.channels[i]
            chan_dict = {}
            chan_dict['data'] = c.data.astype("=f8")
            chan_dict['samples_per_second'] = c.samples_per_second
            chan_dict['name'] = c.name
            chan_dict['frequency_divider'] = c.frequency_divider
            chan_dict['units'] = c.units
            channels[i] = chan_dict
            channel_headers[i] = data.channel_headers[i].data
            channel_dtype_headers[i] = data.channel_dtype_headers[i].data

        d['channels'] = channels

        if not self.data_only:
            d['event_markers'] = self.__build_markers(data)

        d['headers'] = {
            'graph': data.graph_header.data,
            'foreign': data.foreign_header.data,
            'channel': channel_headers,
            'channel_dtype': channel_dtype_headers}

        return d

    def __build_markers(self, data):
        markers = np.zeros(len(data.event_markers), dtype='O')
        for i, marker in enumerate(data.event_markers):
            md = {
                'label': marker.text,
                'sample_index': marker.sample_index,
                'type_code': marker.type_code or 'None',
                'type': marker.type or 'None',
                'date_created': marker.date_created_str,
                'channel_number': marker.channel_number or -1,
                'channel': marker.channel_name or 'Global'
            }
            markers[i] = md
        return markers



#!/usr/bin/env python
# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread
#
# This script pulls all the markers from an AcqKnowledge file and writes it
# to a delimited format.

"""Print the event markers from an AcqKnowledge file.

Usage:
  acq_markers [options] <file>...
  acq_markers -h | --help
  acq_markers --version

Options:
  -o <file>     Write to a file instead of standard output.
"""

from __future__ import unicode_literals, absolute_import, division, with_statement

import sys
import csv

from docopt import docopt
from bioread import reader
from bioread import _metadata as meta


FIELDS = [
    "filename",
    "time (s)",
    "label",
    "channel",
    "date_created",
    "type_code",
    "type",
]


def u8fx():
    if isinstance("x", str):
        return lambda s: s
    else:
        return lambda s: s.encode("utf-8")


uf = u8fx()


def marker_formatter(acq_filename, graph_sample_msec):
    """Return a function that turns a marker into a dict."""

    def f(marker):
        return {
            "filename": uf(acq_filename),
            "time (s)": (marker.sample_index * graph_sample_msec) / 1000,
            "label": uf(marker.text),
            "channel": uf(marker.channel_name or "Global"),
            "date_created": uf(marker.date_created_str),
            "type_code": uf(marker.type_code or "None"),
            "type": uf(marker.type),
        }

    return f


def acq_markers_output_file(input_filenames, output_filename):
    with open(output_filename, "w") as f:
        return acq_markers(input_filenames, f)


def acq_markers(input_filenames, output_stream=None):
    """out示范输出
    [{'filename': '1103-6-活动没启动.acq', 'time (s)': 148.288, 'label': 'Segment 1',
    'channel': 'Global', 'date_created': '2023-11-03T07:11:38   .905000+00:00',
    'type_code': 'apnd', 'type': 'Append'}, ...]
    其中时间为ISO 8601格式，可以datetime.fromisoformat(time_str1)转换为datetime格式"""
    if output_stream is None:
        out = []
    else:
        out = csv.DictWriter(output_stream, FIELDS, delimiter=str("\t"))
        out.writeheader()
    for fname in input_filenames:
        with open(fname, "rb") as infile:
            r = reader.Reader.read_headers(infile)
            mf = marker_formatter(fname, r.graph_header.sample_time)
            for m in r.datafile.event_markers:
                if output_stream is None:
                    out.append(mf(m))
                else:
                    out.writerow(mf(m))
    if output_stream is None:
        return out


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    pargs = docopt(__doc__, args, version=meta.version_description)

    if pargs["-o"]:
        return acq_markers_output_file(pargs["<file>"], pargs["-o"])
    else:
        return acq_markers(pargs["<file>"], sys.stdout)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python
# coding: utf8

from setuptools import setup, find_packages
import sys
import os

SETUP_REQUIRES = ['setuptools >= 30.3.0']
SETUP_REQUIRES += ['wheel'] if 'bdist_wheel' in sys.argv else []


def get_locals(filename):
    l = {}
    exec(open(filename, 'r').read(), {}, l)
    return l


metadata = get_locals(os.path.join('bioread', '_metadata.py'))

setup(
    name="meng-bioread",
    setup_requires=SETUP_REQUIRES,
    version=metadata['version'],
    packages=find_packages(),
    keywords="science research physiological biopac convert library",
    description='A modified version of bioread'
)
